// (c) 2014 Frank Bartels (frank@toppoint.de)
// Licence (GPL3): see LICENCE.txt

#include "LightMatrixKernelLib.h"

#include "PixelStrip.h"

// Defaults: Dimming using poti on pin A3, powered by 14/16
// Override in Animation Code if you like (set in user_init())
int PotiPower= 1;
int PotiPower0= 14;
int PotiPower1= 16;
int PotiBrightness= A3;
int PotiUser0= 15; // A1

int bright_shift=3; // max= 0:maxcurrent 1:maxcurrent/2 2:maxcurrent/4 3:maxcurrent/8
                    // maxcurrent typically <= NPixels * 3*20mA
int bright_shift_max=3; // Poti limit, <=bright_shift_min , max the smaller value (results in larger currents)!
int bright_shift_min=3; // Poti limit

int pinSCK=13;
int pinData=11;
int pinLE=10;
int pinOE=9;

int triangles =0;
int simsquares=1;

int pinDebug=13; // -1=not debug, 4=typical pin


const int NPixels = (SCRdx*SCRdy);

// Output Pin 4 is hardcoded for ATmega328...!
PixelStrip strip = PixelStrip(NPixels);

typedef struct rgb_struct
{ unsigned char r,g,b;
} rgb;
rgb return_rgb;

unsigned char scr[BUFs][SCRdx][SCRdy]; // (real) screen
int act_buf_read=BUFs-1;
int act_buf_write=0;

int pin_pwm_timer  = 3;  // DON'T CHANGE ! (internal)
int int_pwm_timer  = 1;  // DON'T CHANGE ! (internal)
int frq_pwm_select = -2;  // DON'T CHANGE ! (internal)   -1=500Hz
// -2:250Hz -1:500Hz 0:1kHz 
// dont use here: 1:2kHz(4/2) 2:max(32kHz/?) refresh   // full refresh: x/cdz/PWMsteps (/9/5: )

// frq_pwm_select->frq_pwm_select_clockinc : 0->1 -1->2 -2->4
int frq_pwm_select_clockinc= 1;  // 1 is dummy; really calculated in init() ! as that:
  // frq_pwm_select_clockinc= (1<<(0-frq_pwm_select))*time_turbo;

void set_grey_bit(unsigned char i, unsigned char lum)
{ // Register Version
  //int pinSCK=13;  PB5
  //int pinData=11; PB3
  
  if (i > lum) PORTB |= 0x08; // Data=1
  else         PORTB &= 0xF7; // Data=0
  PORTB |= 0x20;   // SCK=1
  PORTB &= 0xDF;   // SCK=0
  /* // Arduino Version 
   i=i > lum ? HIGH : LOW;
   digitalWrite(pinData, i);
   digitalWrite(pinSCK, HIGH);
   digitalWrite(pinSCK, LOW);
  */
}

void to_special_mode(int special) // 0:normal 1:special
{
  digitalWrite(pinData, LOW);
  digitalWrite(pinLE, LOW);
  digitalWrite(pinSCK, LOW);

  digitalWrite(pinOE, HIGH); // OE 1
  frame_delay(1);
  digitalWrite(pinSCK, HIGH);
  frame_delay(1);
  digitalWrite(pinSCK, LOW);
  frame_delay(1);

  digitalWrite(pinOE, LOW); // OE 0
  frame_delay(1);
  digitalWrite(pinSCK, HIGH);
  frame_delay(1);
  digitalWrite(pinSCK, LOW);
  frame_delay(1);

  digitalWrite(pinOE, HIGH); // OE 1
  frame_delay(1);
  digitalWrite(pinSCK, HIGH);
  frame_delay(1);
  digitalWrite(pinSCK, LOW);
  frame_delay(1);

  if (special) digitalWrite(pinLE, HIGH);
  digitalWrite(pinOE, HIGH); // OE 1
  frame_delay(1);
  digitalWrite(pinSCK, HIGH);
  frame_delay(1);
  digitalWrite(pinSCK, LOW);
  frame_delay(1);

  digitalWrite(pinLE, LOW);
  digitalWrite(pinOE, HIGH); // OE 1
  frame_delay(1);
  digitalWrite(pinSCK, HIGH);
  frame_delay(1);
  digitalWrite(pinSCK, LOW);
  frame_delay(1);
}

int display_refresh_running=0;
volatile int clock=0;
unsigned char lumsqrt[16];
void lumsqrt_init()
{ for (int i=0; i<16; i++) lumsqrt[i]=(unsigned char)(sqrt(i*14)*16); // 0 .. 231.86
}
void display_refresh()
{ 
  static int lum=0;
  static int lum2=0;

  if (pinDebug==4) PORTD |= 0x10; // 00010000 Debug=1
  else digitalWrite(pinDebug, HIGH);

  //clock+=(1-frq_pwm_select);  // works for frq_pwm_select = 0 and -1
  clock+=frq_pwm_select_clockinc;
  if (clock>=0x4000) clock=0;
  
if (pinDebug==4) PORTD &= 0xEF; // 00010000 Debug=1
else digitalWrite(pinDebug, LOW);

return; // skip old display procedure!


  if (display_refresh_running) return;
  display_refresh_running=1;

  lum+=96;  // OK: 68, 63,49 ,48+
  //lum+=48;  // OK: 68, 63,49 ,48+
  if (lum>=256) lum-=256; // oder mit 255...

  //lum2= lumsqrt[lum>>4];  // version with gamma
  lum2= lum;              // version without gamma

  //lum2= (int)(sqrt(lum)*16);
  //lum2= lum;
  int lum2g=lum2>>2;
  int lum2r=lum2>>5;

  //              0:&=          1:|=
  // SCK =13; PB5 0xDF 11011111 0x20 00100000
  // Data=11; PB3 0xF7 11110111 0x08 00001000
  // LE  =10; PB2 0xFB 11111011 0x04 00000100
  // OE  = 9; PB1 0xFD 11111101 0x02 00000010

  PORTB &= 0xDF;
  //digitalWrite(pinSCK, LOW);
  PORTB &= 0xF7;
  //digitalWrite(pinData, LOW);
  PORTB &= 0xFB;
  //digitalWrite(pinLE, LOW);
  PORTB &= 0xFD;
  //digitalWrite(pinOE, LOW);
  
  unsigned char r,g,b;
  unsigned char rgb;
  for (int x=0; x<SCRdx; x++)
  { unsigned char *scr_y=&scr[act_buf_read][x][SCRdy-1];
    //for (int y=0; y<SCRdy; y++)
    for (int y=0; y<SCRdy; y++, scr_y--)
    { rgb= *scr_y; // rgb bits: bbgggrrr
    
      b= rgb & 0xC0; // 11000000
      set_grey_bit(b,lum2);
      g= rgb & 0x38; // 00111000
      set_grey_bit(g,lum2g);
      r= rgb & 0x07; // 00000111
      set_grey_bit(r,lum2r);
    }
  }
  //PORTB |= 0x20;   // Farben werden hier verkorkst...
  //digitalWrite(pinSCK, LOW);  // undo SCK=HIGH at end
  
  PORTB |= 0x04;
  PORTB |= 0x04;
  PORTB |= 0x04;
  PORTB |= 0x04;  // repeat same to wait a bit (one cmd = 2clocks)
  //digitalWrite(pinLE, HIGH);
  
  PORTB &= 0xFB;
  //digitalWrite(pinLE, LOW);
  
  display_refresh_running=0;

}

int clear_rgb=0x00; // clear color bits: bbgggrrr

int show_x=0;
int show_y=0;
char ascii_light[]=" .-:!i*oxxxxxxxx00000000########??";
void show_as_ascii(int lines, int chars) // if lines!=0 then lines is used
{ if (lines) chars=lines*SCRdx;
  for(; chars>0; chars--)
  { int rgb=scr[act_buf_read][show_x][show_y];
    int b= (rgb & 0xC0)>>5; // 11000000
    int g= (rgb & 0x38)>>2; // 00111000 (>>3,*2)
    int r= (rgb & 0x07);    // 00000111
    int pix=(b+g+r);
    Serial.print(ascii_light[pix]);
    /*
    if (scr[act_buf_read][show_x][show_y])
      Serial.print('_');
    else
      Serial.print('#');
    */
    show_x++;
    if (show_x>=SCRdx)
    { Serial.print((char)13); // ggf. CR
      Serial.print((char)10); // LF
      show_x=0;
      show_y++;
      if (show_y>=SCRdy)
      { Serial.print((char)13); 
        Serial.print((char)10); 
        for (int j=0; j<SCRdy; j++)
        { Serial.print((char)27); 
          Serial.print("[1;1f"); 
        }
        show_y=0; 
        return; 
      }
    }
  }
}

void display_bufferxada()
{ unsigned char r,g,b,pixel;
  int d,dd,toggle,j=0;
  static RGBPixel p;

  if (!triangles)  // std matrix
  { for (int y=SCRdy-1; y>=0; )
    { for (int x=0; x<SCRdx; x++)
      { pixel= scr[act_buf_read][x][y]; // ?255:0;
  //  scr[act_buf_write][x][y]= (b&0xc0) | ((g&0xe0)>>2) | ((r&0xe0)>>5);
        r=(pixel<<5);
        r=r>>bright_shift;
        g=(pixel<<2)&0xe0;
        g=g>>bright_shift;
        b=pixel&0xc0;
        b=b>>bright_shift;
        p.red   = g;
        p.green = r;
        p.blue  = b;
        strip.setPixelColor(j++, p);
      }
      y--;
      for (int x=SCRdx-1; x>=0; x--)
      { pixel= scr[act_buf_read][x][y]; // ?255:0;
        r=(pixel<<5);
        r=r>>bright_shift;
        g=(pixel<<2)&0xe0;
        g=g>>bright_shift;
        b=pixel&0xc0;
        b=b>>bright_shift;
        p.red   = g;
        p.green = r;
        p.blue  = b;
        strip.setPixelColor(j++, p);
      }
      y--;
    }
  }
  else if (triangles && simsquares)
  { for (int j=SCRdx*SCRdy-1; j>=0; j--)
      strip.setPixelColor(j, 0,0,0);
    d=1; dd=-1;
    toggle=0;
    j=0;
    for (int y=SCRdy-1; y>=SCRdy*0+0; )
    { int jstart=j;
      // 0. same dir, 0:full 1:parts etc.
      for (int x=0; x<SCRdx; )
      { pixel= scr[act_buf_read][toggle?SCRdx-1-x:x][y]; // ?255:0;
        r=(pixel<<5);
        r=r>>bright_shift;
        g=(pixel<<2)&0xe0;
        g=g>>bright_shift;
        b=pixel&0xc0;
        b=b>>bright_shift;
        if (x>=d && x<SCRdx-d)
        { strip.addPixelColor(j, g/2,r/2,b/2);
          j++;
        }
        x++; 
        if (x>=SCRdx) break;

        
        pixel= scr[act_buf_read][toggle?SCRdx-1-x:x][y]; // ?255:0;
        r=(pixel<<5);
        r=r>>bright_shift;
        g=(pixel<<2)&0xe0;
        g=g>>bright_shift;
        b=pixel&0xc0;
        b=b>>bright_shift;
        if (x>      d+1)      strip.addPixelColor(j-1, g/4,r/4,b/4);
        if (x<SCRdx-d-1)      strip.addPixelColor(j+1, g/4,r/4,b/4);
        if (x>=d && x<SCRdx-d) 
        { strip.addPixelColor(j,                       g/3,r/3,b/3);
            //g/4+(g/2-g/4-g/8-g/8), r/4+(r/2-r/4-r/8-r/8), b/4+(b/2-b/4-b/8-b/8));
          j++;
        }
        x++; 
      }
      j=jstart;
      y--;
      // 1. y-1, same dir, 0:parts 1:full etc.
      for (int x=0; x<SCRdx; )
      { pixel= scr[act_buf_read][toggle?SCRdx-1-x:x][y]; // ?255:0;
        r=(pixel<<5);
        r=r>>bright_shift;
        g=(pixel<<2)&0xe0;
        g=g>>bright_shift;
        b=pixel&0xc0;
        b=b>>bright_shift;
        if (x>      d+1)       strip.addPixelColor(j-1, g/4,r/4,b/4);
        if (x<SCRdx-d-1)       strip.addPixelColor(j+1, g/4,r/4,b/4);
        if (x>=d && x<SCRdx-d)
        { strip.addPixelColor(j,                        g/3,r/3,b/3);
            //g/4+(g/2-g/4-g/8-g/8), r/4+(r/2-r/4-r/8-r/8), b/4+(b/2-b/4-b/8-b/8));
          j++;
        }
        x++;
        if (x>=SCRdx) break;
        
        pixel= scr[act_buf_read][toggle?SCRdx-1-x:x][y]; // ?255:0;
        r=(pixel<<5);
        r=r>>bright_shift;
        g=(pixel<<2)&0xe0;
        g=g>>bright_shift;
        b=pixel&0xc0;
        b=b>>bright_shift;
        if (x>=d && x<SCRdx-d)
        { strip.addPixelColor(j, g/2,r/2,b/2);
          j++;
        }
        x++;
      }
      y--;
      d+=dd;
      if (d<0) { d=0; dd=1; }
      toggle=!toggle;
    }
  }
  else // triangles && !simsquares
  { // not yet
  }
  
  strip.show();

  if (pinDebug==4) PORTD |= 0x10; // 00010000 Debug=1
  else digitalWrite(pinDebug, HIGH);

}

void swap_buffers(int mode) // mode: 0=Swap 1=SwapAndClear 2=Copy
{
//scr[0][0][0]=32;
//  display_bufferxx();  
//  return;
 //while (display_refresh_running) { /* wait */ } // step is optional against flickering


  act_buf_read=0;
  act_buf_write=0;

  //display_bufferxx();  
  display_bufferxada();  

  if (DO_show_as_ascii) show_as_ascii(80,0);

  if (mode==1)
  { for(int y=0; y<SCRdy; y++)
      for(int x=0; x<SCRdx; x++)
        scr[act_buf_write][x][y]=clear_rgb;
  }

  /*
  if (mode!=2)
    if (act_buf_read)
    { 
      act_buf_read=0;
      act_buf_write=1;
    }
    else
    { 
      act_buf_read=1;
      act_buf_write=0;
    }
  if (mode==1)
  { for(int y=0; y<SCRdy; y++)
      for(int x=0; x<SCRdx; x++)
        scr[act_buf_write][x][y]=clear_rgb;
  }
  else if (mode==2)
  { for(int y=0; y<SCRdy; y++)
      for(int x=0; x<SCRdx; x++)
        scr[act_buf_read][x][y]=scr[act_buf_write][x][y];
  }*/

  if (PotiPower)
    bright_shift= (1023-analogRead(PotiBrightness)) *(bright_shift_min-bright_shift_max+1)/1024 +bright_shift_max;
}

void time_turbo_update()
{ frq_pwm_select_clockinc= (1<<(0-frq_pwm_select))*time_turbo;
}

void setup()
{ // Internal stuff
  time_turbo_update();
  lumsqrt_init();

  // System stuff
  strip.begin();
  strip.show();

  user_init();  // No output here! (paint and the like)

  pinMode(pinDebug, OUTPUT);
  digitalWrite(pinDebug, HIGH);
  pinMode(pinSCK, OUTPUT);
  pinMode(pinData, OUTPUT);
  pinMode(pinLE, OUTPUT);
  pinMode(pinOE, OUTPUT);
  to_special_mode(0);
  swap_buffers(1);
  swap_buffers(1);
  // more system stuff - must be at end (start display_refresh())
  pinMode(        pin_pwm_timer, OUTPUT);
  analogWrite(    pin_pwm_timer, 32);
  attachInterrupt(int_pwm_timer, display_refresh, RISING);
  TCCR2B = TCCR2B & 0b11111000 | (3-frq_pwm_select); // -1=500Hz
  // TCCR2B = TCCR2B & 0b11111000 | (frq_pwm_select==0?3:(frq_pwm_select==1?2:1)); // 3906Hz/976Hz - http://www.arduino.cc/playground/Main/TimerPWMCheatsheet
  digitalWrite(pinDebug, LOW);
  digitalWrite(pinDebug, HIGH);

  if (PotiPower)
  { pinMode(PotiPower0,OUTPUT); digitalWrite(PotiPower0,0);
    pinMode(PotiPower1,OUTPUT); digitalWrite(PotiPower1,1);
    bright_shift= (1023-analogRead(PotiBrightness)) *(bright_shift_min-bright_shift_max+1)/1024 +bright_shift_max;
  }

  // must be at end of setup (trashes TCCR2B and vice versa)
  if (DO_show_as_ascii)
    Serial.begin(115200); // default 9600, 38400, 115200
}

volatile int target_clock=0;
void frame_delay(int ms)
{
  target_clock+=ms*slowmotion;

  volatile int dummy=0x7fff;
  int clock0=clock;
  if (clock+0x2000<target_clock)
    target_clock-=0x4000;
  if (clock>=target_clock) target_clock=clock; 
  else 
    for (; clock<target_clock;)
    { if (clock<clock0)
      { clock0=clock;
        target_clock-=0x4000;
      }
      if (dummy) dummy--;
      else if (clock==0) return; // safety exit
    }
}


// Current font has only: SPACE and A-Z (capital letters) !!

#define CFONTwidth  0x5 // (font constant)
#define CFONTstart 0x40
#define CFONTend   0x5b
//const unsigned char cfont[] __attribute__ ((section (".const"))) =
//const unsigned char cfont[] __attribute__((progmem)) =
unsigned char cfont[]=
//const prog_uint8_t cfont[] =   // convert ptr with... pgm_read_byte
{ /*
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x20
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x21
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x22
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x23
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x24
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x25
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x26
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x27
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x28
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x29
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x2a
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x2b
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x2c
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x2d
  0x00, 0x00, 0x20, 0x00, 0x00, // 0x2e
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x2f

  //0x1e, 0x31, 0x29, 0x25, 0x1e, // 0x30 0
  0x1e, 0x21, 0x2d, 0x21, 0x1e, // 0x30 0
  0x00, 0x22, 0x3f, 0x20, 0x00, // 0x31 1
  0x22, 0x31, 0x29, 0x29, 0x26, // 0x32 2
  0x12, 0x21, 0x25, 0x25, 0x1a, // 0x33 3
  0x18, 0x14, 0x12, 0x3f, 0x10, // 0x34 4
  0x22, 0x25, 0x25, 0x25, 0x18, // 0x35 5
  //0x1e, 0x29, 0x29, 0x29, 0x1a, // 0x36 6
  0x1e, 0x25, 0x25, 0x25, 0x18, // 0x36 6
  0x01, 0x01, 0x31, 0x0d, 0x03, // 0x37 7
  0x1a, 0x25, 0x25, 0x25, 0x1a, // 0x38 8
  0x06, 0x29, 0x29, 0x29, 0x1e, // 0x39 9
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x3a
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x3b
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x3c
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x3d
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x3e
  0x00, 0x00, 0x00, 0x00, 0x00, // 0x3f
  */

  0x00, 0x00, 0x00, 0x00, 0x00, // SPACE (0x40 or 0x20)
  0x38, 0x16, 0x11, 0x16, 0x38, // A     (0x41)
  0x3f, 0x25, 0x25, 0x26, 0x18, // B     (0x42)
  0x1e, 0x21, 0x21, 0x21, 0x12, // C
  0x3f, 0x21, 0x21, 0x21, 0x1e, // D
  0x3f, 0x25, 0x25, 0x25, 0x21, // E
  0x3f, 0x05, 0x05, 0x05, 0x01, // F
  0x1e, 0x21, 0x21, 0x29, 0x1a, // G
  0x3f, 0x08, 0x08, 0x08, 0x3f, // H
  0x00, 0x21, 0x3f, 0x21, 0x00, // I
  0x00, 0x11, 0x21, 0x1f, 0x00, // J
  //0x3f, 0x08, 0x14, 0x14, 0x22, // K
  0x3f, 0x08, 0x14, 0x12, 0x21, // K
  0x3f, 0x20, 0x20, 0x20, 0x00, // L
  0x3f, 0x02, 0x0c, 0x02, 0x3f, // M
  0x3f, 0x02, 0x04, 0x08, 0x3f, // N
  0x1e, 0x21, 0x21, 0x21, 0x1e, // O
  0x3f, 0x09, 0x09, 0x09, 0x06, // P
  0x1e, 0x21, 0x29, 0x11, 0x2e, // Q
  0x3f, 0x09, 0x09, 0x19, 0x26, // R
  0x22, 0x25, 0x25, 0x25, 0x18, // S
  0x01, 0x01, 0x3f, 0x01, 0x01, // T
  0x1f, 0x20, 0x20, 0x20, 0x1f, // U
  0x03, 0x0c, 0x30, 0x0c, 0x03, // V
  0x3f, 0x10, 0x0c, 0x10, 0x3f, // W
  0x31, 0x0a, 0x04, 0x0a, 0x31, // X
  0x01, 0x02, 0x3c, 0x02, 0x01, // Y
  0x31, 0x29, 0x25, 0x23, 0x21, // Z
};

//#define CFONTwidth  0x5 // (font constant)
//#define CFONTstart 0x20
//#define CFONTend   0x5b
//unsigned char  cfont[]=

void char_paint(int x, int y, char c, 
                   unsigned char r, unsigned char g, unsigned char b)
{ if (c<CFONTstart || c>=CFONTend) return;
  int color=get_index(r,g,b);
  int pos_in_font=(c-CFONTstart)*5;
  for(int xx=0; xx<CFONTwidth; xx++) if (x+xx>=0 && x+xx<SCRdx)
  { //unsigned char col=pgm_read_byte(&cfont[pos_in_font+xx]);
    unsigned char col=cfont[pos_in_font+xx]; 
    for(int yy=0; yy<8; yy++)
    { if (col&1)
        if (y+yy>=0 && y+yy<SCRdy)
          scr[act_buf_write][x+xx][y+yy]=color;
      col>>=1;
    }
  }
}
// x,y:       left/top
// show:      print the chars slowly one after the other
// scrolling: scroll text if too large
const int CFONTwidthPlus=(CFONTwidth+2);
void string_paint(int xx, int y, char *s, int show, int scrolling,
                     unsigned char r, unsigned char g, unsigned char b)
{ int len=strlen(s)*(CFONTwidth+2);
  int toolarge=len-SCRdx;
  if (!scrolling || toolarge<0) toolarge=0;
  for (int scroll=0; scroll<=toolarge; scroll++) 
  { if (scroll==1) swap_buffers(1);
    for (int i=0, x=xx;  s[i]!=0;  i++, x+=CFONTwidthPlus)
    { char_paint(x-scroll,y,s[i],r,g,b);
      if (scroll==0 && show) { swap_buffers(2); if (x<SCRdx) frame_delay(120); }
    }
    if (scroll>0) { frame_delay(50); swap_buffers(1); }
  }
}
void string_anim(char *s, int dt, int show, 
                    unsigned char r, unsigned char g, unsigned char b)
{ swap_buffers(1);
  string_paint(0,1,s,show,1,r,g,b);
  //swap_buffers(1);
  frame_delay(dt);
}

// hi versions
void char_paint_hi(int x, int y, char c, 
                   int r, int g, int b)
{ if (c<CFONTstart || c>=CFONTend) return;
  int pos_in_font=(c-CFONTstart)*5;
  for(int xx=0; xx<CFONTwidth; xx++) if (x+xx*0x100>=0 && x+xx*0x100<SCRdx*0x100)
  { //unsigned char col=pgm_read_byte(cfont[pos_in_font+xx]);
    unsigned char col=cfont[pos_in_font+xx]; 
    for(int yy=0; yy<8; yy++)
    { if (col&1)
        if (y+yy*0x100>=0 && y+yy*0x100<SCRdy*0x100)
          paint_hi(x+xx*0x100,y+yy*0x100, r,g,b);
          //scr[act_buf_write][x+xx][y+yy]=color;
      col>>=1;
    }
  }
}
// x,y:       left/top
// show:      print the chars slowly one after the other
// scrolling: scroll text if too large
void string_paint_hi(int xx, int y, char *s, int show, int scrolling,
                     int r, int g, int b)
{ unsigned int len=strlen(s)*(CFONTwidth+1);
  if (len>0xff) { string_paint(xx>>8,y>>8,s,show,scrolling,r,g,b); return; } // would overflow
       //    void string_paint(int xx, int y, char *s, int show, int scrolling,
       //            unsigned char r, unsigned char g, unsigned char b)
  len*=0x100;
  unsigned int toolarge=len-SCRdx*0x100;
  unsigned int scrolldx=0x50;
  unsigned int scroll0=0;
  if (!scrolling || toolarge<0) toolarge=0;
  for (int scroll=scroll0; scroll<=toolarge; scroll+=scrolldx)
  { if (scroll==0) swap_buffers(1);
    for (unsigned int i=0, x=xx;  s[i]!=0;  i++, x+=(CFONTwidth+1)*0x100)
    { char_paint_hi(x-scroll,y, s[i], r,g,b);
      if (scroll==scroll0 && show) { swap_buffers(2); if ((x>>8)<SCRdx) frame_delay(200); }
    }
    //if (scroll>=scrolldx) { frame_delay(250); swap_buffers(1); }
    frame_delay(40); swap_buffers(1);
  }
}
void string_anim_hi(char *s, int dt, int show, 
                       int r, int g, int b)
{ swap_buffers(1);
  string_paint_hi(0,0x80,s,show,1,r,g,b);
  //swap_buffers(1);
  frame_delay(dt);
}
void paint(int x, int y, unsigned int r, unsigned int g, unsigned int b)
{ if (x<0 || y<0 || x>=SCRdx || y>=SCRdy) return;
  if (r>255) r=255;
  if (g>255) g=255;
  if (b>255) b=255;

  scr[act_buf_write][x][y]= (b&0xc0) | ((g&0xe0)>>2) | ((r&0xe0)>>5);
}
void paint_add(int x, int y, int r, int g, int b)
{ if (x<0 || y<0 || x>=SCRdx || y>=SCRdy) return;

  if (r<0x10 && g<0x10 & b<0x10) return;

  unsigned int old= scr[act_buf_write][x][y];
  unsigned int rr= (old<<4)&0x70;
  rr+=(r>>1); if (rr>0x70) rr=0x70;
  unsigned int gg= (old<<1)&0x70;
  gg+=(g>>1); if (gg>0x70) gg=0x70;
  unsigned int bb= (old>>1)&0x60;
  bb+=(b>>1); if (bb>0x60) bb=0x60;
  scr[act_buf_write][x][y]=  ((bb&0x60)<<1) + ((gg&0x70)>>1) + ((rr&0x70)>>4);
}

void get_pixel(int x, int y)
{ if (x<0 || y<0 || x>=SCRdx || y>=SCRdy) return;
  unsigned char pix= scr[act_buf_write][x][y];
  return_rgb.r= (pix<<5)&0xe0;
  return_rgb.g= (pix<<2)&0xe0;
  return_rgb.b= (pix   )&0xc0;
  //return return_rgb;
}
unsigned char get_index(unsigned char r, unsigned char g, unsigned char b)
{ return((b&0xc0) | ((g&0xe0)>>2) | (r>>5));
}
void paint_index(int x, int y, unsigned char rgb)
{ if (x<0 || y<0 || x>=SCRdx || y>=SCRdy) return;
  scr[act_buf_write][x][y]= rgb;
}

void paint_hi(int x, int y, int r, int g, int b)
// x,y contain fixed point factor FIXP
// r,b,g<0x1000 (=16*256) (but normal brighness <=255)
{ if (r<0) r=0;
  if (g<0) g=0;
  if (b<0) b=0;
  
  x-=FIXP/2;
  y-=FIXP/2;
  
  int xr=(x+FIXP/2)&(FIXP-1);
  x=(x+FIXP/2)>>FIXbit;
  int yr=(y+FIXP/2)&(FIXP-1);
  y=(y+FIXP/2)>>FIXbit;

  // lum = max(r,g,b)
  int l=(r>g?(r>b?r:b):(b>g?b:g))>>4;  // l<0x1000 >>4 -> <0x100
  // factor d (<2**8)
  int dd0=(0xff-xr)>>1; // <0x80
  int dd1=(xr)>>1;
  int d[2][2];
  d[0][0]=((0xff-yr)*dd0)>>8; // <0x80 * <0x100 / 0x100 -> <0x80
  d[1][0]=((0xff-yr)*dd1)>>8;
  d[0][1]=(yr*dd0)>>8;
  d[1][1]=(yr*dd1)>>8;
  int e[4][4];
  for (int yy=0; yy<4; yy++)
    for (int xx=0; xx<4; xx++)
      e[xx][yy]=0;
  for (int yy=1; yy<3; yy++)
    for (int xx=1; xx<3; xx++)
    { int v=(d[xx-1][yy-1]*l)>>3;   // <0x100 * <0x80 >>3 -> <0x1000  (norm1=0xff)
      int toomuch=v-0x100; if (toomuch<0) toomuch=0;
      v-=toomuch;
      int r=toomuch&3;
      toomuch>>2;
      e[xx][yy]+=v;
      e[xx-1][yy]+=toomuch;
      e[xx+1][yy]+=toomuch;
      e[xx][yy-1]+=toomuch;
      e[xx][yy+1]+=toomuch;
      if (r>0) e[xx-1][yy-1]+=1;
      if (r>1) e[xx+1][yy-1]+=1;
      if (r>2) e[xx+1][yy+1]+=1;
      //for (int yyy=-1; yyy<=1; yyy++)
      //  for (int xxx=-1; xxx<=1; xxx++) // if (xxx!=0 || yyy!=0)  (eigentlich /8 aber egal...)
      //    e[xx+xxx][yy+yyy]+=toomuch;
    }
  for (int yy=0; yy<4; yy++)
    for (int xx=0; xx<4; xx++) if (e[xx][yy])
    { int lf=(e[xx][yy]>>4);
      int rr=(r>>4)*lf;
      if (rr>0xff) rr=0xff;
      int gg=(g>>4)*lf;
      if (gg>0xff) gg=0xff;
      int bb=(b>>4)*lf;
      if (bb>0xff) bb=0xff;
      paint_add(x+xx-1,y+yy-1, rr,gg,bb);
    }
}


